#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Clément Warneys <clement.warneys@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This inner package is overloading the original **watchdog** package 
by revising and completing it, resolving the current situation
where the useful **watchdog** package is not maintained anymore...

Please read original **watchdog** project documentation for more information:
https://pypi.org/project/watchdog/

revised_watchdog.events
=======================

.. automodule:: lazydog.revised_watchdog.events
   :members:



revised_watchdog.observers.inotify
==================================

.. automodule:: lazydog.revised_watchdog.observers.inotify
   :members:

revised_watchdog.observers.inotify_c
====================================

.. automodule:: lazydog.revised_watchdog.observers.inotify_c
   :members:
   
revised_watchdog.observers.inotify_buffer
=========================================

.. automodule:: lazydog.revised_watchdog.observers.inotify_buffer
   :members:

"""